# Copyright (c) Alibaba, Inc. and its affiliates.
NON_ENTITY_LABEL = 'O'
NON_ENTITY_LABEL_ID = 0
PAD_LABEL = 'X'
PAD_LABEL_ID = -100

OBJECT_START_TOKEN = '<O>'
SUBJECT_START_TOKEN = '<E>'
NONE_REL_LABEL = 'None'
NONE_REL_LABEL_ID = 0

PARTIAL_LABEL = 'P'
PARTIAL_LABEL_ID = -1
