# Copyright (c) Alibaba, Inc. and its affiliates.
# trigger register mechanism
from adaseq import data, metrics, models, ms_patch, pipelines, training
from adaseq.utils.logging import prepare_global_logging
from adaseq.version import __version__

prepare_global_logging()
